/*

  NRF24L01+

     http://aterlux.ru/index.php?page=article&art=nrf24l01p
, !  !

 250/
        0      CRC16

 DEBUG_NRF24     .
 DEBUG_NRF24_REGISTERS      print_registers.

(c) sslobodyan@yandex.ru

2016

*/

#include <Arduino.h> 
#include <SPI.h>
#include "NRF24.h"

//#define DEBUG_NRF24
//#define DEBUG_NRF24_REGISTERS

#if defined (__AVR__)
	#define spi_nrf SPI			//   
#elif defined (__STM32F1__)
	SPIClass spi_nrf(1);		//    SPI  (2)
#endif
 
//      print_registers 
#define prdebug Serial.print
#define prdebugln Serial.println

//     
#ifdef DEBUG_NRF24
	#define debug Serial.print
	#define debugln Serial.println
#else	
	#define debug //
	#define debugln //
#endif

//    
NRF24::NRF24(uint8_t _cepin, uint8_t _cspin)
{
	cepin = _cepin;
	cspin = _cspin;
    pinMode(cepin, OUTPUT);
    pinMode(cspin, OUTPUT);
	csn_deassert();
	deassert_ce();
}

//        
NRF24::NRF24(uint8_t _cepin, uint8_t _cspin, uint8_t _irqpin)
{
	irqpin = _irqpin;
	pinMode(irqpin, INPUT);
	use_irq_pin = true;
    NRF24(_cepin, _cspin);
}

//    1   SPI,   
uint8_t NRF24::spi_send_recv(uint8_t data) {
  return spi_nrf.transfer(data);
}

//    ( )   CE
inline void NRF24::assert_ce() {
	digitalWrite(cepin, HIGH);
}

//    ( )   CE
inline void NRF24::deassert_ce() {
	digitalWrite(cepin, LOW);
}

//    ( )   CSN
inline void NRF24::csn_assert() {
	digitalWrite(cspin, LOW);
}

//    ( )   CSN
inline void NRF24::csn_deassert() {
	digitalWrite(cspin, HIGH);
}

//   cmd,   count  ,     buf,   
uint8_t NRF24::read_buf(uint8_t cmd, uint8_t * buf, uint8_t count) {
	uint8_t tmp;
  csn_assert();
  uint8_t status = spi_send_recv(cmd);
  debug(F("read_buf cmd="));  debug(cmd,HEX);  debug(" -> ");
  while (count--) {
	tmp = spi_send_recv(0xFF);
	debug(tmp,HEX);	debug(",");
    *(buf++) = tmp;
  }
  debug(" done ");
  csn_deassert();
  return status;
}

//   cmd,   count     buf,   
uint8_t NRF24::write_buf(uint8_t cmd, uint8_t * buf, uint8_t count) {
  csn_assert();
  uint8_t status = spi_send_recv(cmd);
  debug("write_buf cmd=");
  debug(cmd,HEX);
  debug(" -> ");
  while (count--) {
	debug(*buf,HEX);  
	debug(",");
    spi_send_recv(*(buf++));
  }
  debugln(" done");
  csn_deassert();
  return status;
}

//     reg ( 0  31)   
uint8_t NRF24::readreg(uint8_t reg) {
  csn_assert();
  spi_send_recv((reg & 31) | R_REGISTER);
  uint8_t answ = spi_send_recv(0xFF);
  csn_deassert();
  return answ;
}

//     reg ( 0  31),   
uint8_t NRF24::writereg(uint8_t reg, uint8_t val) {
  csn_assert();
  uint8_t status = spi_send_recv((reg & 31) | W_REGISTER);
  spi_send_recv(val);
  csn_deassert();
  return status;
}

//  count    reg ( 0  31)      buf,
//   
uint8_t NRF24::readreg_buf(uint8_t reg, uint8_t * buf, uint8_t count) {
  return read_buf((reg & 31) | R_REGISTER, buf, count);
}

//  count    buf    reg ( 0  31),   
uint8_t NRF24::writereg_buf(uint8_t reg, uint8_t * buf, uint8_t count) {
  return write_buf((reg & 31) | W_REGISTER, buf, count);
}

//      FIFO  
uint8_t NRF24::read_rx_payload_width() {
  csn_assert();
  spi_send_recv(R_RX_PL_WID);
  uint8_t answ = spi_send_recv(0xFF);
  csn_deassert();
  return answ;
}

//  .   
uint8_t NRF24::cmd(uint8_t cmd) {
  csn_assert();
  uint8_t status = spi_send_recv(cmd);
  csn_deassert();
  return status;
}

//  1,    IRQ  () .
uint8_t NRF24::is_interrupt_with_irq_pin() {
  return digitalRead(irqpin);
}

uint8_t NRF24::is_interrupt_no_pin() {
//       !!!
  return (cmd(NOP) & ((1 << RX_DR) | (1 << TX_DS) | (1 << MAX_RT))) ? 1 : 0;
}

//     (     -    )
uint8_t NRF24::is_interrupt() {
	if ( use_irq_pin ) return is_interrupt_with_irq_pin();
	else return is_interrupt_no_pin();
}


//    BK
void NRF24::activate(uint8_t code)
{
  csn_assert();
  SPI.transfer(ACTIVATE);
  SPI.transfer(code);
  csn_deassert();
}


//     .  1,   , 0   
uint8_t NRF24::begin() {

  debugln("begin");

  deassert_ce();

#if defined (__AVR__)
    spi_nrf.begin();
    spi_nrf.setClockDivider(SPI_CLOCK_DIV2); // 16/2 = 8
    spi_nrf.setDataMode(SPI_MODE0);
    spi_nrf.setBitOrder(MSBFIRST);
#elif defined (__STM32F1__)
    spi_nrf.begin();
    spi_nrf.setClockDivider(SPI_CLOCK_DIV8); // 72/8 = 9
    spi_nrf.setDataMode(SPI_MODE0);
    spi_nrf.setBitOrder(MSBFIRST);
#endif

  for(uint8_t cnt = 100;;) {
    writereg(CONFIG, (1 << EN_CRC) | (1 << CRCO) | (1 << PRIM_RX)); //  
    if (readreg(CONFIG) == ((1 << EN_CRC) | (1 << CRCO) | (1 << PRIM_RX))) 
      break;
    //      ,    -  ,   .
    if (!cnt--)
      return 0; //   100      ,    
      for (byte i=0; i<200; i++) delay_15us();
  }

  //writereg(EN_AA, (1 << ENAA_P1)); //      1
  writereg(EN_AA, 0); //   
  writereg(EN_RXADDR, (1 << ERX_P0) ); //   0 
  writereg(SETUP_RETR, SETUP_RETR_DELAY_250MKS | SETUP_RETR_NO_RETRANSMIT); //  
  
  writereg(RF_SETUP, RF_SETUP_250KBPS | RF_SETUP_0DBM); //   250 /   0dBm
  //writereg(RF_SETUP, RF_SETUP_1MBPS | RF_SETUP_0DBM); //   1m/   0dBm
  
  writereg(SETUP_AW, SETUP_AW_5BYTES_ADDRESS); //    5 

  update_local_address();
  update_remote_address();
  update_channel();
  
  writereg(RX_PW_P0, 32); //       0:  1  32. 0 -   .
  
  activate(0x73);

  writereg(FEATURE, 0x05); //        
  writereg(DYNPD, (1 << DPL_P0) ); //      0 
  

  writereg(CONFIG, (1 << EN_CRC) | (1 << CRCO) | (1 << PWR_UP) | (1 << PRIM_RX)); //  
  for (byte i=0; i<100; i++) delay_15us(); //    2 ms
  flushRX(); //   
  flushTX(); //   
  assert_ce();

  return (readreg(CONFIG) == ((1 << EN_CRC) | (1 << CRCO) | (1 << PWR_UP) | (1 << PRIM_RX))) ? 1 : 0;
}

//     
uint8_t NRF24::TX_FIFO_empty() {
	return ( readreg(FIFO_STATUS) & (1 << TX_EMPTY) );
}

//     
uint8_t NRF24::RX_FIFO_empty() {
	return ( readreg(FIFO_STATUS) & (1 << RX_EMPTY));
}

//    
uint8_t NRF24::RX_FIFO_full() {
	return ( readreg(FIFO_STATUS) & (1 << RX_FULL));
}

//    
uint8_t NRF24::TX_FIFO_full() {
	return ( readreg(FIFO_STATUS) & (1 << TX_FULL_FIFO));
}

//    
uint8_t NRF24::flushRX() {
	debugln("FLUSH_RX!");
	return ( cmd(FLUSH_RX) );
}

//    
uint8_t NRF24::flushTX() {
	debugln("FLUSH_TX!");
	return ( cmd(FLUSH_TX) );
}

//   
uint8_t NRF24::update_local_address(){
	return writereg_buf(RX_ADDR_P0, local_address, 5); //    0 	
}

//    
uint8_t NRF24::update_remote_address(){
	return writereg_buf(TX_ADDR, remote_address, 5); 	
}

//    (0-125)
uint8_t NRF24::update_channel(){
	return writereg(RF_CH, channel); 	
}


//        . 
// buf -   , size -   ( 1  32)
uint8_t NRF24::send_data(uint8_t * buf, uint8_t size) {
  debug("Send data ");
  uint8_t conf = readreg(CONFIG);
  if (!(conf & (1 << PWR_UP))) {//    -  ,   
	debugln("Power down!");
    return 0; 
  }
  deassert_ce(); //    ,    
  uint8_t status = writereg(CONFIG, conf & ~(1 << PRIM_RX)); //   
  if (status & (1 << TX_FULL_STATUS)) { //    ,   
	debugln("TX FIFO full!");
	flushTX();
  }
  flushRX(); //   ,       
  write_buf(W_TX_PAYLOAD_NOACK, buf, size); //       
  assert_ce(); //    CE    
  delay_15us(); //        .   10,   
  deassert_ce();
  return 1;
}

//     
// buf -   
//      0
uint8_t NRF24::get_data(uint8_t * buf) {
	debug("get_data ");
	if ( RX_FIFO_empty() ) { // FIFO 
		debugln("RX FIFO empty!");
		return 0;
	}
    uint8_t len = read_rx_payload_width(); //   
    if ( len > 32 ) { //   .   
      cmd(FLUSH_RX); 
	  debugln("Bad RX length!");
	  return 0;
    }
	read_buf(R_RX_PAYLOAD, buf, len); //  
	debug("=> ");
	debug(len);
	debugln(" bytes");
	return len;
}


//   
void NRF24::listening() {
	uint8_t conf = readreg(CONFIG);
	writereg(CONFIG, conf | (1 << PRIM_RX)); //   PRIM_RX: 
	assert_ce(); //     CE  -   
}


//       
//    -         TX_handler
//     ,      rx_buffer      RX_handler
//     -   ,     
void NRF24::check() {
    uint8_t status = cmd(NOP);
    writereg(STATUS, status); //    ,     
	uint8_t fifo_status = readreg(FIFO_STATUS);
    if (status & (1 << RX_DR) ) { //  
		while ( ! (fifo_status & (1 << RX_EMPTY)) ) {
			uint8_t len = get_data(rx_buffer); //     
			if ( RX_handler )  {
				(*RX_handler)(len); //     
			}
			fifo_status = readreg(FIFO_STATUS);
		}
    }
    if (status & (1 << TX_DS) ) { //  
		listening();
		if ( TX_handler )  {
			(*TX_handler)(); //     
		}
    }	
	if ( go_tx ) { //     
		go_tx = false;
		send_data(tx_buffer, tx_buffer_len);
	}
	if ( fifo_status & (1 << RX_FULL) ) flushRX();
}


//   15       
#if defined (__AVR__)
	void NRF24::delay_15us(){ //  16 
		for (byte n=0; n<40; n++) {
			asm("nop;nop;nop;nop;");
		}
	}
#elif defined (__STM32F1__)
	void NRF24::delay_15us(){ //  72 STM32
		for (byte n=0; n<150; n++) asm("nop;");
	}
#endif


#ifndef DEBUG_NRF24_REGISTERS
	void NRF24::print_registers(void){};
#else
//     
//     
void NRF24::print_registers(void){

	uint8_t reg1, i;
	uint8_t reg5[5];
	
	prdebugln(""); prdebugln(F("NRF2401 Registers:"));
	
	reg1 = readreg(CONFIG);
	prdebug(F("CONFIG=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t  "));
	prdebug(F(" MASK_RX_DR="));prdebug(reg1&(1<<MASK_RX_DR)?1:0);
	prdebug(F(" MASK_TX_DS="));prdebug(reg1&(1<<MASK_TX_DS)?1:0);
	prdebug(F(" MASK_MAX_RT="));prdebug(reg1&(1<<MASK_MAX_RT)?1:0);
	prdebug(F(" EN_CRC="));prdebug(reg1&(1<<EN_CRC)?F("1 "):F("0  CRC"));
	prdebug(F(" CRCO="));prdebug(reg1&(1<<CRCO)?F("1 16 "):F("0 8 "));
	prdebug(F(" PWR_UP="));prdebug(reg1&(1<<PWR_UP)?1:0);
	prdebug(F(" PRIM_RX="));prdebug(reg1&(1<<PRIM_RX)?F("1 "):F("0 "));prdebugln("");
	
	reg1 = readreg(EN_AA);
	prdebug(F("EN_AA=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t    "));
	prdebug(F(" ENAA_P5="));prdebug(reg1&(1<<ENAA_P5)?1:0);
	prdebug(F(" ENAA_P4="));prdebug(reg1&(1<<ENAA_P4)?1:0);
	prdebug(F(" ENAA_P3="));prdebug(reg1&(1<<ENAA_P3)?1:0);
	prdebug(F(" ENAA_P2="));prdebug(reg1&(1<<ENAA_P2)?1:0);
	prdebug(F(" ENAA_P1="));prdebug(reg1&(1<<ENAA_P1)?1:0);
	prdebug(F(" ENAA_P0="));prdebug(reg1&(1<<ENAA_P0)?1:0);prdebugln("");
	
	reg1 = readreg(EN_RXADDR);
	prdebug(F("EN_RXADDR=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t    "));
	prdebug(F(" ERX_P5="));prdebug(reg1&(1<<ERX_P5)?1:0);
	prdebug(F(" ERX_P4="));prdebug(reg1&(1<<ERX_P4)?1:0);
	prdebug(F(" ERX_P3="));prdebug(reg1&(1<<ERX_P3)?1:0);
	prdebug(F(" ERX_P2="));prdebug(reg1&(1<<ERX_P2)?1:0);
	prdebug(F(" ERX_P1="));prdebug(reg1&(1<<ERX_P1)?1:0);
	prdebug(F(" ERX_P0="));prdebug(reg1&(1<<ERX_P0)?1:0);prdebugln("");

	reg1 = readreg(SETUP_AW);
	prdebug(F("SETUP_AW=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t   "));
	switch(reg1) {
		case 1: prdebug(F(" 3 "));break;
		case 2: prdebug(F(" 4 "));break;
		case 3: prdebug(F(" 5 "));break;
		default: prdebug(F(" ???"));
	}
	prdebugln("");

	reg1 = readreg(SETUP_RETR);
	prdebug(F("SETUP_RETR=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t   "));
	prdebug(F(" ARD=0x"));prdebug(reg1>>4,HEX);prdebug(" ");prdebug(250*(reg1>>4));prdebug(F(" mks "));
	prdebug(F(" ARC=0x"));prdebug(reg1&0x0F,HEX);prdebug(" ");prdebug(reg1&0x0F);prdebugln(F("  "));
	
	reg1 = readreg(RF_CH);
	prdebug(F("RF_CH=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t  "));prdebugln(reg1);
	
	reg1 = readreg(RF_SETUP);
	prdebug(F("RF_SETUP=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t   "));
	prdebug(F(" CONT_WAVE="));prdebug(reg1&(1<<CONT_WAVE)?1:0);
	prdebug(F(" RF_DR_LOW="));prdebug(reg1&(1<<RF_DR_LOW)?F("1  250\"):F("0  \"));
	prdebug(F(" PLL_LOCK="));prdebug(reg1&(1<<PLL_LOCK)?1:0);
	prdebug(F(" RF_DR_HIGH="));prdebug(reg1&(1<<RF_DR_HIGH)?F("1 2\"):F("0 1\"));
	prdebug(F(" RF_PWR="));prdebug((reg1&(0b11<<RF_PWR))>>RF_PWR);
	switch ((reg1&(0b11<<RF_PWR))>>RF_PWR) {
		case 0: prdebug(F(" -18dBm"));break;
		case 1: prdebug(F(" -12dBm"));break;
		case 2: prdebug(F(" -6dBm"));break;
		case 3: prdebug(F(" 0dBm"));break;
	}
	prdebugln("");

	reg1 = readreg(STATUS);
	prdebug(F("STATUS=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t  "));
	prdebug(F(" RX_DR="));prdebug(reg1&(1<<RX_DR)?1:0);
	prdebug(F(" TX_DS="));prdebug(reg1&(1<<TX_DS)?1:0);
	prdebug(F(" MAX_RT="));prdebug(reg1&(1<<MAX_RT)?1:0);
	prdebug(F(" TX_FULL="));prdebug(reg1&(1<<TX_FULL_STATUS)?1:0);
	prdebug(F(" RX_P_NO="));prdebug((reg1&(0b111<<RX_P_NO))>>RX_P_NO);prdebugln(F("  FIFO RX"));

	reg1 = readreg(OBSERVE_TX);
	prdebug(F("OBSERVE_TX=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t   "));
	prdebug(F(" PLOS_CNT=0x"));prdebug(reg1>>4,HEX);prdebug(" ");prdebug(reg1>>4);prdebug(F("  "));
	prdebug(F(" ARC_CNT=0x"));prdebug(reg1&0x0F,HEX);prdebug(" ");prdebug(reg1&0x0F);prdebugln(F("    "));
	
	reg1 = readreg(RPD);
	prdebug(F("RPD=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t   -64dBm "));prdebugln(reg1>0?F(" "):F(" "));
	
	readreg_buf(RX_ADDR_P0, reg5, 5);
	prdebug(F("RX_ADDR_P0=0x"));for(i=0;i<5;i++) prdebug(reg5[i],HEX);prdebugln(F("\t   0 "));
	
	readreg_buf(RX_ADDR_P1, reg5, 5);
	prdebug(F("RX_ADDR_P1=0x"));for(i=0;i<5;i++) prdebug(reg5[i],HEX);prdebugln(F("\t   1 "));
	
	reg1 = readreg(RX_ADDR_P2);
	prdebug(F("RX_ADDR_P2=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebugln(F("\t   2 "));

	reg1 = readreg(RX_ADDR_P3);
	prdebug(F("RX_ADDR_P3=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebugln(F("\t   3 "));

	reg1 = readreg(RX_ADDR_P4);
	prdebug(F("RX_ADDR_P4=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebugln(F("\t   4 "));

	reg1 = readreg(RX_ADDR_P5);
	prdebug(F("RX_ADDR_P5=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebugln(F("\t   5 "));

	readreg_buf(TX_ADDR, reg5, 5);
	prdebug(F("TX_ADDR=0x"));for(i=0;i<5;i++) prdebug(reg5[i],HEX);prdebugln(F("\t   "));

	reg1 = readreg(RX_PW_P0);
	prdebug(F("RX_PW_P0=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t    0 ="));prdebugln(reg1);
	
	reg1 = readreg(RX_PW_P1);
	prdebug(F("RX_PW_P1=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t    1 ="));prdebugln(reg1);
	
	reg1 = readreg(RX_PW_P2);
	prdebug(F("RX_PW_P2=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t    2 ="));prdebugln(reg1);
	
	reg1 = readreg(RX_PW_P3);
	prdebug(F("RX_PW_P3=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t    3 ="));prdebugln(reg1);
	
	reg1 = readreg(RX_PW_P4);
	prdebug(F("RX_PW_P4=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t    4 ="));prdebugln(reg1);
	
	reg1 = readreg(RX_PW_P5);
	prdebug(F("RX_PW_P5=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t    5 ="));prdebugln(reg1);
	
	reg1 = readreg(FIFO_STATUS);
	prdebug(F("FIFO_STATUS=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t   "));
	prdebug(F(" TX_REUSE="));prdebug(reg1&(1<<TX_REUSE)?1:0);
	prdebug(F(" TX_FULL="));prdebug(reg1&(1<<TX_FULL_FIFO)?1:0);
	prdebug(F(" TX_EMPTY="));prdebug(reg1&(1<<TX_EMPTY)?1:0);
	prdebug(F(" RX_FULL="));prdebug(reg1&(1<<RX_FULL)?1:0);
	prdebug(F(" RX_EMPTY="));prdebug(reg1&(1<<RX_EMPTY)?1:0);prdebugln("");
	
	reg1 = readreg(DYNPD);
	prdebug(F("DYNPD=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t     "));
	prdebug(F(" DPL_P5="));prdebug(reg1&(1<<DPL_P5)?1:0);
	prdebug(F(" DPL_P4="));prdebug(reg1&(1<<DPL_P4)?1:0);
	prdebug(F(" DPL_P3="));prdebug(reg1&(1<<DPL_P3)?1:0);
	prdebug(F(" DPL_P2="));prdebug(reg1&(1<<DPL_P2)?1:0);
	prdebug(F(" DPL_P1="));prdebug(reg1&(1<<DPL_P1)?1:0);
	prdebug(F(" DPL_P0="));prdebug(reg1&(1<<DPL_P0)?1:0);prdebugln("");
	
	reg1 = readreg(FEATURE);
	prdebug(F("FEATURE=0x"));prdebug(reg1>0x0F?"":"0");prdebug(reg1,HEX);prdebug(F("\t   "));
	prdebug(F(" EN_DPL="));prdebug(reg1&(1<<EN_DPL)?F("1   "):F("0   ,"));
	prdebug(F(" EN_ACK_PAY="));prdebug(reg1&(1<<EN_ACK_PAY)?F("1     "):F("0     ,"));
	prdebug(F(" EN_DYN_ACK="));prdebug(reg1&(1<<EN_DYN_ACK)?F("1    "):F("0    "));
	prdebugln("");
	
			
}
#endif



